#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2015 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest
from typing import Any
from queue import SimpleQueue
from collections.abc import Callable

from xpra.gtk.gobject import one_arg_signal
from xpra.net.protocol.socket_handler import SocketProtocol
from xpra.net.subprocess_wrapper import SubprocessCaller, SubprocessCallee
from xpra.net.bytestreams import Connection
from xpra.net.common import PacketType
from xpra.os_util import gi_import
from xpra.log import Logger

log = Logger("test")

TEST_TIMEOUT = 10*1000

GObject = gi_import("GObject")
GLib = gi_import("GLib")


class FakeSubprocess:
    """ defined just so the protocol layer can call terminate() on it """
    def __init__(self):
        self.returncode: int | None = None

    def terminate(self) -> None:
        self.returncode = 1

    def poll(self) -> int | None:
        return self.returncode


class LoopbackConnection(Connection):
    """ a fake connection which just writes back whatever is sent to it """
    def __init__(self, *args):
        Connection.__init__(self, *args)
        self.queue = SimpleQueue()

    def read(self, _n):
        self.may_abort("read")
        # FIXME: we don't handle n...
        return self.queue.get(True)

    def write(self, buf, packet_type: str = "") -> int:
        self.may_abort("write")
        self.queue.put(buf)
        return len(buf)

    def close(self) -> None:
        self.queue.put(None)
        Connection.close(self)

    def may_abort(self, _action) -> None:
        assert self.active


def loopback_protocol(process_packet_cb: Callable[[Any, PacketType], None],
                      get_packet_cb: Callable[[], [PacketType, bool, bool]]) -> SocketProtocol:
    conn = LoopbackConnection("fake", "fake")
    protocol = SocketProtocol(conn, process_packet_cb, get_packet_cb=get_packet_cb)
    protocol.enable_encoder("rencodeplus")
    protocol.enable_compressor("none")
    return protocol


class LoopbackProcess(SubprocessCaller):
    """ a fake subprocess which uses the loopback connection """
    def exec_subprocess(self):
        return FakeSubprocess()

    def make_protocol(self) -> SocketProtocol:
        return loopback_protocol(self.process_packet, self.get_packet)


class LoopbackCallee(SubprocessCallee):

    def make_protocol(self) -> SocketProtocol:
        return loopback_protocol(self.process_packet, self.get_packet)


class TestCallee(GObject.GObject):
    __gsignals__ = {
        "test-signal": one_arg_signal,
    }


GObject.type_register(TestCallee)


class SubprocessWrapperTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        unittest.TestCase.setUpClass()
        from xpra.net import packet_encoding
        packet_encoding.init_all()
        from xpra.net import compression
        compression.init_compressors("none")

    def test_loopback_caller(self) -> None:
        mainloop = GLib.MainLoop()
        lp = LoopbackProcess()
        readback = []

        def record_packet(_lp, *args) -> None:
            readback.append(args)
        lp.connect("foo", record_packet)

        def stop() -> None:
            # this may deadlock on win32..
            lp.stop_protocol()
            lp.stop_process()
            GLib.idle_add(mainloop.quit)

        def end(*_args) -> None:
            stop()

        lp.connect("end", end)
        self.timeout = False

        def timeout_error() -> None:
            self.timeout = True
            stop()
        GLib.timeout_add(TEST_TIMEOUT, timeout_error)
        sent_str = b"hello foo"
        GLib.idle_add(lp.send, "foo", sent_str)
        GLib.idle_add(lp.send, "bar", b"hello bar")
        GLib.idle_add(lp.send, "end")
        lp.stop = stop
        # run!
        lp.start()
        mainloop.run()
        assert len(readback)==1, "expected 1 record in loopback but got %s" % len(readback)
        rss = readback[0][0]
        assert rss== sent_str, "expected message string '%s' but got '%s'" % (sent_str, rss)
        assert self.timeout is False, "the test did not exit cleanly (not received the 'end' packet?)"

    def test_loopback_callee(self) -> None:
        mainloop = GLib.MainLoop()
        callee = TestCallee()
        lc = LoopbackCallee(wrapped_object=callee, method_whitelist=["test_signal", "loop_stop", "unused"])
        # this will cause the "test-signal" to be sent via the loopback connection
        lc.connect_export("test-signal")
        readback = []

        def test_signal_function(*args) -> None:
            log("test_signal_function%s", args)
            readback.append(args)
            GLib.idle_add(lc.send, "loop_stop")
        # hook up a function which will be called when the wrapper converts the packet into a method call:
        callee.test_signal = test_signal_function
        # lc.connect_export("test-signal", hello)
        self.timeout = False
        callee.timeout_fn = None

        def loop_stop(*args) -> None:
            log("loop_stop%s timeout_fn=%s", args, callee.timeout_fn)
            if callee.timeout_fn:
                GLib.source_remove(callee.timeout_fn)
                callee.timeout_fn = None
            lc.stop()
            GLib.idle_add(mainloop.quit)

        def timeout_error() -> None:
            log.warn("timeout_error")
            callee.timeout_fn = None
            self.timeout = True
            loop_stop()
        callee.timeout_fn = GLib.timeout_add(TEST_TIMEOUT, timeout_error)
        signal_string = b"hello foo"
        GLib.idle_add(callee.emit, "test-signal", signal_string)
        # hook up a stop function call which ends this test cleanly
        callee.loop_stop = loop_stop
        # run!
        lc.start()
        mainloop.run()
        lc.stop()
        log("readback=%s", readback)
        assert len(readback) == 1, "expected 1 record in loopback but got %s" % len(readback)
        rss = readback[0][0]
        log("rss=%s, timeout=%s", rss, self.timeout)
        assert rss == signal_string, "expected signal string '%s' but got '%s'" % (signal_string, rss)
        assert self.timeout is False, "the test did not exit cleanly (not received the 'end' packet?)"


def main():
    unittest.main()


if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from struct import Struct
from enum import IntEnum

# merge header and packet if packet is smaller than:
PIXEL_FORMAT: bytes = b"BBBBHHHBBBBBB"
CLIENT_INIT: bytes = b"!HH" + PIXEL_FORMAT + b"I"


class RFBClientMessage(IntEnum):
    """ client to server messages """
    SetPixelFormat = 0
    SetEncodings = 2
    FramebufferUpdateRequest = 3
    KeyEvent = 4
    PointerEvent = 5
    ClientCutText = 6
    # optional:
    FileTransfer = 7
    SetScale = 8
    SetServerInput = 9
    SetSW = 10
    TextChat = 11
    KeyFrameRequest = 12
    KeepAlive = 13
    SetScaleFactor = 15
    RequestSession = 20
    SetSession = 21
    NotifyPluginStreaming = 80
    VMWare = 127
    CarConnectivity = 128
    EnableContinuousUpdates = 150
    ClientFence = 248
    OliveCallControl = 249
    XVPClientMessage = 250
    SetDesktopSize = 251
    Tight = 252
    GIIClientMessage = 253
    # VMWARE = 254
    QEMUClientMessage = 255


CLIENT_PACKET_TYPE_STR: dict[int, str] = {}
for cm in RFBClientMessage:
    CLIENT_PACKET_TYPE_STR[cm.value] = cm.name

PACKET_STRUCT: dict[int, Struct] = {}
for msg, fmt in {
    RFBClientMessage.SetPixelFormat: b"!BBBB" + PIXEL_FORMAT,
    RFBClientMessage.SetEncodings: b"!BBH",
    RFBClientMessage.FramebufferUpdateRequest: b"!BBHHHH",
    RFBClientMessage.KeyEvent: b"!BBBBi",
    RFBClientMessage.PointerEvent: b"!BBHH",
    RFBClientMessage.ClientCutText: b"!BBBBi",
}.items():
    PACKET_STRUCT[msg] = Struct(fmt)


class RFBServerMessage(IntEnum):
    # server to client messages:
    FRAMEBUFFERUPDATE = 0
    SETCOLORMAPENTRIES = 1
    BELL = 2
    SERVERCUTTEXT = 3
    # optional:
    RESIZEFRAMEBUFFER1 = 4
    KEYFRAMEUPDATE = 4
    FILETRANSFER = 7
    TEXTCHAT = 11
    KEEPALIVE = 13
    RESIZEFRAMEBUFFER2 = 15
    VMWARE1 = 127
    CARCONNECTIVITY = 128
    ENDOFCONTINOUSUPDATES = 150
    SERVERSTATE = 173
    SERVERFENCE = 248
    OLIVECALLCONTROL = 249
    XVPSERVERMESSAGE = 250
    TIGHT = 252
    GIISERVERMESSAGE = 253
    VMWARE2 = 254
    QEMUSERVERMESSAGE = 255


SERVER_PACKET_TYPE_STR: dict[int, str] = {}
for sm in RFBServerMessage:
    SERVER_PACKET_TYPE_STR[sm.value] = sm.name


class RFBEncoding(IntEnum):
    RAW = 0
    COPYRECT = 1
    RRE = 2
    CORRE = 4
    HEXTILE = 5
    ZLIB = 6
    TIGHT = 7
    ZLIBHEX = 8
    TRLE = 15
    ZRLE = 16
    H264 = 20
    JPEG = 21
    JRLE = 22
    VAH264 = 23
    ZRLE2 = 24
    HITACHI_ZYWRLE = 17
    DESKTOPSIZE = -223
    LASTRECT = -224
    CURSOR = -239
    XCURSOR = -240
    QEMU_POINTER = -257
    QEMU_KEY = -258
    QEMU_AUDIO = -259
    TIGHT_PNG = -260
    GII = -305
    DESKTOPNAME = -307
    EXTENDEDDESKTOPSIZE = -308
    XVP = -309
    FENCE = -312
    CONTINUOUSUPDATES = -313
    CURSORWITHALPHA = -314
    VA_H264 = 0x48323634

    # -23 to -32    JPEG Quality Level Pseudo-encoding
    # -247 to -256    Compression Level Pseudo-encoding
    # -412 to -512    JPEG Fine-Grained Quality Level Pseudo-encoding
    # -763 to -768    JPEG Subsampling Level Pseudo-encoding


ENCODING_STR: dict[int, str] = {}
for e in RFBEncoding:
    ENCODING_STR[e.value] = e.name


class RFBAuth(IntEnum):
    INVALID = 0
    NONE = 1
    VNC = 2
    TIGHT = 16
    VeNCrypt = 19


AUTH_STR: dict[int, str] = {
    RFBAuth.INVALID: "Invalid",
    RFBAuth.NONE: "None",
    RFBAuth.VNC: "VNC",
    RFBAuth.TIGHT: "Tight",
    5: "RA2",
    6: "RA2ne",
    17: "Ultra",
    18: "TLS",
    RFBAuth.VeNCrypt: "VeNCrypt",
    20: "SASL",
    21: "MD5",
    22: "xvp",
}
for i in (3, 4):
    AUTH_STR[i] = "RealVNC"
for i in range(7, 16):
    AUTH_STR[i] = "RealVNC"
for i in range(128, 255):
    AUTH_STR[i] = "RealVNC"
for i in range(30, 35):
    AUTH_STR[i] = "Apple"

RFB_KEYNAMES: dict[int, str] = {
    0xff08: "BackSpace",
    0xff09: "Tab",
    0xff0d: "Return",
    0xff1b: "Escape",
    0xff63: "Insert",
    0xffff: "Delete",
    0xff50: "Home",
    0xff57: "End",
    0xff55: "PageUp",
    0xff56: "PageDown",
    0xff51: "Left",
    0xff52: "Up",
    0xff53: "Right",
    0xff54: "Down",
    0xffe1: "Shift_L",
    0xffe2: "Shift_R",
    0xffe3: "Control_L",
    0xffe4: "Control_R",
    0xffe7: "Meta_L",
    0xffe8: "Meta_R",
    0xffe9: "Alt_L",
    0xffea: "Alt_R",
    0x20: "space",
    0x60: "grave",
    0x2d: "minus",
    0x3d: "equal",
    0x2b: "plus",
    0x5f: "underscore",
    0xac: "notsign",
    0x21: "exclam",
    0x22: "quotedbl",
    0x24: "dollar",
    0x5e: "asciicircum",
    0x26: "ampersand",
    0x2a: "asterisk",
    0x28: "parenleft",
    0x29: "parenright",
    0x5c: "backslash",
    0x7c: "bar",
    0x5b: "bracketleft",
    0x5d: "bracketright",
    0x7b: "braceleft",
    0x7d: "braceright",
    0x3b: "semicolon",
    0x27: "apostrophe",
    0x23: "numbersign",
    0x3a: "colon",
    0x40: "at",
    0x7e: "asciitilde",
    0x2c: "comma",
    0x2e: "period",
    0x2f: "slash",
    0x3c: "less",
    0x3e: "greater",
    0x3f: "question",
    0xffb0: "KP_0",
    0xffb1: "KP_1",
    0xffb2: "KP_2",
    0xffb3: "KP_3",
    0xffb4: "KP_4",
    0xffb5: "KP_5",
    0xffb6: "KP_6",
    0xffb7: "KP_7",
    0xffb8: "KP_8",
    0xffb9: "KP_9",
    0xffae: "KP_Decimal",
    0xffaf: "KP_Divide",
    0xffaa: "KP_Multiply",
    0xffad: "KP_Subtract",
    0xffab: "KP_Add",
    0xff8d: "KP_Enter",
    0xff7f: "Num_Lock",
    0xff9f: "KP_Delete",
    0xff9e: "KP_Insert",
    0xff9c: "KP_End",
    0xff9b: "KP_Down",
    # 0xffb3      : "KP_Next",
    0xff96: "KP_Left",
    0xff9d: "KP_Begin",
    0xff98: "KP_Right",
    0xff95: "KP_Home",
    0xff97: "KP_Up",
    0xff9a: "KP_Prior",
    0xff14: "Scroll_Lock",
    0xff13: "Pause",
    0x1008ff26: "XF86Back",
    0x1008ff27: "XF86Forward",
    0x1008ff18: "XF86HomePage",
}

for i in range(1, 13):
    RFB_KEYNAMES[0xffbe + (i - 1)] = "F%i" % i

RFB_KEYS: dict[str, int] = {}
for keysym, name in RFB_KEYNAMES.items():
    RFB_KEYS[name.lower()] = keysym



# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

#cython: wraparound=False

from typing import Any, Dict, Tuple
from collections.abc import Sequence
from time import monotonic

from libc.stdint cimport uintptr_t
from xpra.codecs.image import ImageWrapper
from xpra.buffers.membuf cimport makebuf, MemBuf, buffer_context     # pylint: disable=syntax-error
from xpra.codecs.constants import get_subsampling_divs
from xpra.codecs.constants import VideoSpec
from xpra.codecs.debug import may_save_image
from xpra.net.compression import Compressed
from xpra.util.env import envbool
from xpra.util.objects import typedict
from xpra.util.str_fn import csv
from xpra.log import Logger
log = Logger("encoder", "jpeg")

cdef int YUV = envbool("XPRA_TURBOJPEG_YUV", True)


ctypedef int TJSAMP
ctypedef int TJPF

cdef extern from "math.h":
    double sqrt(double arg) nogil
    double round(double arg) nogil

cdef extern from "turbojpeg.h":
    TJSAMP  TJSAMP_444
    TJSAMP  TJSAMP_422
    TJSAMP  TJSAMP_420
    TJSAMP  TJSAMP_GRAY
    TJSAMP  TJSAMP_440
    TJSAMP  TJSAMP_411

    TJPF    TJPF_RGB
    TJPF    TJPF_BGR
    TJPF    TJPF_RGBX
    TJPF    TJPF_BGRX
    TJPF    TJPF_XBGR
    TJPF    TJPF_XRGB
    TJPF    TJPF_GRAY
    TJPF    TJPF_RGBA
    TJPF    TJPF_BGRA
    TJPF    TJPF_ABGR
    TJPF    TJPF_ARGB
    TJPF    TJPF_CMYK

    int TJFLAG_BOTTOMUP
    int TJFLAG_FASTUPSAMPLE
    int TJFLAG_FASTDCT
    int TJFLAG_ACCURATEDCT

    ctypedef void* tjhandle
    tjhandle tjInitCompress()
    int tjDestroy(tjhandle handle)
    char* tjGetErrorStr()
    #unsigned long tjBufSize(int width, int height, int jpegSubsamp)
    int tjCompress2(tjhandle handle, const unsigned char *srcBuf,
                    int width, int pitch, int height, int pixelFormat, unsigned char **jpegBuf,
                    unsigned long *jpegSize, int jpegSubsamp, int jpegQual, int flags) nogil

    int tjCompressFromYUVPlanes(tjhandle handle,
                    const unsigned char **srcPlanes,
                    int width, const int *strides,
                    int height, int subsamp,
                    unsigned char **jpegBuf,
                    unsigned long *jpegSize, int jpegQual,
                    int flags) nogil

TJPF_VAL: Dict[str, int] = {
    "RGB"   : TJPF_RGB,
    "BGR"   : TJPF_BGR,
    "RGBX"  : TJPF_RGBX,
    "BGRX"  : TJPF_BGRX,
    "XBGR"  : TJPF_XBGR,
    "XRGB"  : TJPF_XRGB,
    "GRAY"  : TJPF_GRAY,
    "RGBA"  : TJPF_RGBA,
    "BGRA"  : TJPF_BGRA,
    "ABGR"  : TJPF_ABGR,
    "ARGB"  : TJPF_ARGB,
    "CMYK"  : TJPF_CMYK,
}
TJSAMP_STR: Dict[int, str] = {
    TJSAMP_444  : "444",
    TJSAMP_422  : "422",
    TJSAMP_420  : "420",
    TJSAMP_GRAY : "GRAY",
    TJSAMP_440  : "440",
    TJSAMP_411  : "411",
}


def get_version() -> Tuple[int, int]:
    return (2, 0)


def get_type() -> str:
    return "jpeg"


def get_info() -> Dict[str, Any]:
    return {"version"   : get_version()}


def get_encodings() -> Sequence[str]:
    return ("jpeg", "jpega")


if YUV:
    JPEG_INPUT_COLORSPACES = ("BGRX", "RGBX", "XBGR", "XRGB", "RGB", "BGR", "YUV420P", "YUV422P", "YUV444P")
else:
    JPEG_INPUT_COLORSPACES = ("BGRX", "RGBX", "XBGR", "XRGB", "RGB", "BGR")
JPEGA_INPUT_COLORSPACES = ("BGRA", "RGBA", )


def get_specs() -> Sequence[VideoSpec]:
    specs: Sequence[VideoSpec] = []
    for encoding in ("jpeg", "jpega"):
        in_css = JPEG_INPUT_COLORSPACES if encoding == "jpeg" else JPEGA_INPUT_COLORSPACES

        for in_cs in in_css:
            width_mask=0xFFFF
            height_mask=0xFFFF
            if in_cs in ("YUV420P", "YUV422P"):
                width_mask=0xFFFE
            if in_cs in ("YUV420P", ):
                height_mask=0xFFFE

            specs.append(VideoSpec(
                    encoding=encoding, input_colorspace=in_cs, output_colorspaces=(in_cs, ),
                    has_lossless_mode=False,
                    codec_class=Encoder, codec_type="jpeg",
                    setup_cost=0, cpu_cost=100, gpu_cost=0,
                    min_w=16, min_h=16, max_w=16*1024, max_h=16*1024,
                    can_scale=False,
                    score_boost=-50,
                    width_mask=width_mask, height_mask=height_mask,
                )
            )
    return specs


cdef inline int norm_quality(int quality) nogil:
    if quality<=0:
        return 0
    if quality>=100:
        return 100
    return <int> round(sqrt(<double> quality)*10)


cdef class Encoder:
    cdef tjhandle compressor
    cdef int width
    cdef int height
    cdef object encoding
    cdef object src_format
    cdef int quality
    cdef int grayscale
    cdef long frames
    cdef object __weakref__

    def __init__(self):
        self.width = self.height = self.quality = self.frames = 0
        self.compressor = tjInitCompress()
        if self.compressor == NULL:
            raise RuntimeError("Error: failed to instantiate a JPEG compressor")

    def init_context(self, encoding: str, width : int, height : int, src_format: str, options: typedict) -> None:
        assert encoding in ("jpeg", "jpega"), "invalid encoding: %s" % encoding
        if encoding == "jpeg":
            assert src_format in JPEG_INPUT_COLORSPACES
        elif encoding == "jpega":
            assert src_format in JPEGA_INPUT_COLORSPACES
        else:
            raise ValueError(f"invalid encoding {encoding!r}")
        scaled_width = options.intget("scaled-width", width)
        scaled_height = options.intget("scaled-height", height)
        assert scaled_width == width and scaled_height == height, "jpeg encoder does not handle scaling"
        self.encoding = encoding
        self.width = width
        self.height = height
        self.src_format = src_format
        self.grayscale = options.boolget("grayscale")
        self.quality = options.intget("quality", 50)

    def is_ready(self) -> bool:
        return self.compressor!=NULL

    def is_closed(self) -> bool:
        return self.compressor == NULL

    def clean(self) -> None:
        self.width = self.height = self.quality = 0
        if self.compressor:
            r = tjDestroy(self.compressor)
            self.compressor = NULL
            if r:
                log.error("Error: failed to destroy the JPEG compressor, code %i:", r)
                log.error(" %s", get_error_str())

    def get_encoding(self) -> str:
        return self.encoding

    def get_width(self) -> int:
        return self.width

    def get_height(self) -> int:
        return self.height

    def get_type(self) -> str:
        return "jpeg"

    def get_src_format(self) -> str:
        return self.src_format

    def get_info(self) -> Dict[str,Any]:
        info = get_info()
        info |= {
            "frames"        : int(self.frames),
            "encoding"      : self.encoding,
            "width"         : self.width,
            "height"        : self.height,
            "quality"       : self.quality,
            "grayscale"     : bool(self.grayscale),
        }
        return info

    def compress_image(self, image: ImageWrapper, options: typedict) -> Tuple:
        quality = options.get("quality", -1)
        if quality>0:
            self.quality = quality
        else:
            quality = self.quality
        pfstr = image.get_pixel_format()
        if pfstr in ("YUV420P", "YUV422P", "YUV444P"):
            cdata = encode_yuv(self.compressor, image, quality, self.grayscale)
        else:
            cdata = encode_rgb(self.compressor, image, quality, self.grayscale)
        if not cdata:
            return None
        now = monotonic()
        may_save_image("jpeg", cdata, now)
        client_options = {
            "full-range": image.get_full_range(),
        }
        if self.encoding == "jpega":
            from xpra.codecs.argb.argb import alpha
            a = alpha(image)
            planes = (a, )
            rowstrides = (image.get_rowstride()//4, )
            adata = do_encode_yuv(self.compressor, "YUV400P", planes,
                                  self.width, self.height, rowstrides,
                                  quality, TJSAMP_GRAY)
            client_options["alpha-offset"] = len(cdata)
            may_save_image("jpeg", adata, now)
            cdata = memoryview(cdata).tobytes()+memoryview(adata).tobytes()
        self.frames += 1
        return memoryview(cdata), client_options


def get_error_str() -> str:
    err = tjGetErrorStr()
    try:
        return err.decode("latin1")
    except:
        return str(err)


JPEG_INPUT_FORMATS = ("RGB", "RGBX", "BGRX", "XBGR", "XRGB", )
JPEGA_INPUT_FORMATS = ("RGBA", "BGRA", "ABGR", "ARGB")


def encode(coding, image: ImageWrapper, options: typedict) -> Tuple:
    assert coding in ("jpeg", "jpega")
    rgb_format = image.get_pixel_format()
    if coding == "jpega" and rgb_format.find("A")<0:
        #why did we select 'jpega' then!?
        coding = "jpeg"
    cdef int quality = options.intget("quality", 50)
    cdef int grayscale = options.boolget("grayscale", False)
    cdef int width = image.get_width()
    cdef int height = image.get_height()
    cdef int scaled_width = options.intget("scaled-width", width)
    cdef int scaled_height = options.intget("scaled-height", height)
    cdef char resize = scaled_width!=width or scaled_height!=height
    log("encode%s", (coding, image, options))
    input_formats = JPEG_INPUT_FORMATS if coding == "jpeg" else JPEGA_INPUT_FORMATS
    if rgb_format not in input_formats or resize and len(rgb_format)!=4:
        from xpra.codecs.argb.argb import argb_swap
        if not argb_swap(image, input_formats):
            log("jpeg: argb_swap failed to convert %s to a suitable format: %s" % (
                rgb_format, input_formats))
        log("jpeg converted %s to %s", rgb_format, image)

    if resize:
        from xpra.codecs.argb.scale import scale_image
        image = scale_image(image, scaled_width, scaled_height)
        log("jpeg scaled image: %s", image)

    client_options = {
        "quality"   : quality
    }
    cdef tjhandle compressor = tjInitCompress()
    if compressor == NULL:
        log.error("Error: failed to instantiate a JPEG compressor")
        return ()
    cdef int r
    try:
        cdata = encode_rgb(compressor, image, quality, grayscale)
        if not cdata:
            return None
        now = monotonic()
        may_save_image("jpeg", cdata, now)
        bpp = 24
        if coding == "jpega":
            from xpra.codecs.argb.argb import alpha
            a = alpha(image)
            planes = (a, )
            rowstrides = (image.get_rowstride()//4, )
            adata = do_encode_yuv(compressor, "YUV400P", planes,
                                  width, height, rowstrides,
                                  quality, TJSAMP_GRAY)
            may_save_image("jpeg", adata, now)
            client_options["alpha-offset"] = len(cdata)
            cdata = memoryview(cdata).tobytes()+memoryview(adata).tobytes()
            bpp = 32
        return coding, Compressed(coding, memoryview(cdata), False), client_options, width, height, 0, bpp
    finally:
        r = tjDestroy(compressor)
        if r:
            log.error("Error: failed to destroy the JPEG compressor, code %i:", r)
            log.error(" %s", get_error_str())


cdef inline TJSAMP get_subsamp(int quality):
    if quality<60:
        return TJSAMP_420
    elif quality<80:
        return TJSAMP_422
    return TJSAMP_444


cdef object encode_rgb(tjhandle compressor, image, int quality, int grayscale=0):
    pfstr = image.get_pixel_format()
    pf = TJPF_VAL.get(pfstr)
    if pf is None:
        raise ValueError(f"invalid pixel format {pfstr!r}")
    cdef TJPF tjpf = pf
    cdef TJSAMP subsamp
    if grayscale:
        subsamp = TJSAMP_GRAY
    else:
        subsamp = get_subsamp(quality)
    cdef int width = image.get_width()
    cdef int height = image.get_height()
    cdef int stride = image.get_rowstride()
    pixels = image.get_pixels()
    return do_encode_rgb(compressor, pfstr, pixels,
                         width, height, stride,
                         quality, tjpf, subsamp)


cdef object do_encode_rgb(tjhandle compressor, pfstr, pixels,
                   int width, int height, int stride,
                   int quality, TJPF tjpf, TJSAMP subsamp):
    cdef int flags = 0
    cdef unsigned char *out = NULL
    cdef unsigned long out_size = 0
    cdef int r = -1
    cdef const unsigned char *src
    log("jpeg.encode_rgb with subsampling=%s for pixel format=%s with quality=%s",
        TJSAMP_STR.get(subsamp, subsamp), pfstr, quality)
    with buffer_context(pixels) as bc:
        assert len(bc)>=stride*height, "%s buffer is too small: %i bytes, %ix%i=%i bytes required" % (
            pfstr, len(bc), stride, height, stride*height)
        src = <const unsigned char *> (<uintptr_t> int(bc))
        if src == NULL:
            raise ValueError("missing pixel buffer address from context %s" % bc)
        with nogil:
            r = tjCompress2(compressor, src,
                            width, stride, height, tjpf,
                            &out, &out_size, subsamp, norm_quality(quality), flags)
    if r!=0:
        log.error("Error: failed to compress jpeg image, code %i:", r)
        log.error(" %s", get_error_str())
        log.error(" width=%i, stride=%i, height=%i", width, stride, height)
        log.error(" quality=%i (from %i), flags=%x", norm_quality(quality), quality, flags)
        log.error(" pixel format=%s", pfstr)
        return None
    assert out_size>0 and out!=NULL, "jpeg compression produced no data"
    log("output is %s bytes", out_size)
    return makebuf(out, out_size, 0)


cdef object encode_yuv(tjhandle compressor, image, int quality, int grayscale=0):
    pfstr = image.get_pixel_format()
    assert pfstr in ("YUV420P", "YUV422P"), "invalid yuv pixel format %s" % pfstr
    cdef TJSAMP subsamp
    if grayscale:
        subsamp = TJSAMP_GRAY
    elif pfstr == "YUV420P":
        subsamp = TJSAMP_420
    elif pfstr == "YUV422P":
        subsamp = TJSAMP_422
    elif pfstr == "YUV444P":
        subsamp = TJSAMP_444
    else:
        raise ValueError("invalid yuv pixel format %s" % pfstr)
    cdef int width = image.get_width()
    cdef int height = image.get_height()
    rowstrides = image.get_rowstride()
    planes = image.get_pixels()
    return do_encode_yuv(compressor, pfstr, planes,
                         width, height, rowstrides,
                         quality, subsamp)


cdef object do_encode_yuv(tjhandle compressor, pfstr, planes,
                   int width, int height, rowstrides,
                   int quality, TJSAMP subsamp):
    cdef int flags = 0
    cdef unsigned char *out = NULL
    cdef unsigned long out_size = 0
    cdef int r = -1
    cdef int strides[3]
    cdef const unsigned char *src[3]
    divs = get_subsampling_divs(pfstr)
    for i in range(3):
        src[i] = NULL
        strides[i] = 0
    for i, (xdiv, ydiv) in enumerate(divs):
        assert rowstrides[i]>=width//xdiv, "stride %i is too small for width %i of plane %s from %s" % (
            rowstrides[i], width//xdiv, "YUV"[i], pfstr)
        strides[i] = rowstrides[i]
    contexts = []
    try:
        for i, (xdiv, ydiv) in enumerate(divs):
            bc = buffer_context(planes[i])
            bc.__enter__()
            contexts.append(bc)
            if len(bc)<strides[i]*height//ydiv:
                raise ValueError("plane %r is only %i bytes, %ix%i=%i bytes required" % (
                "YUV"[i], len(bc), strides[i], height, strides[i]*height//ydiv))
            src[i] = <const unsigned char *> (<uintptr_t> int(bc))
            if src[i] == NULL:
                raise ValueError("missing plane %s from context %s" % ("YUV"[i], bc))
        log("jpeg.encode_yuv with subsampling=%s for pixel format=%s with quality=%s",
            TJSAMP_STR.get(subsamp, subsamp), pfstr, quality)
        with nogil:
            r = tjCompressFromYUVPlanes(compressor,
                                        src,
                                        width, <const int*> strides,
                                        height, subsamp,
                                        &out, &out_size, norm_quality(quality), flags)
        if r!=0:
            log.error("Error: failed to compress jpeg image, code %i:", r)
            log.error(" %s", get_error_str())
            log.error(" width=%i, strides=%s, height=%i", width, rowstrides, height)
            log.error(" quality=%i (from %i), flags=%x", norm_quality(quality), quality, flags)
            log.error(" pixel format=%s, subsampling=%s", pfstr, TJSAMP_STR.get(subsamp, subsamp))
            log.error(" planes: %s", csv(<uintptr_t> src[i] for i in range(3)))
            return None
    finally:
        for bc in contexts:
            bc.__exit__()
    assert out_size>0 and out!=NULL, "jpeg compression produced no data"
    return makebuf(out, out_size, 0)


def selftest(full=False) -> None:
    log("jpeg selftest")
    from xpra.codecs.checks import make_test_image
    for q in (0, 50, 100):
        for encoding in ("jpeg", "jpega"):
            img = make_test_image("BGRA", 32, 32)
            v = encode(encoding, img, typedict({"quality" : q}))
            assert v, "encode output was empty!"
            from xpra.util.str_fn import hexstr



# This file is part of Xpra.
# Copyright (C) 2022 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Tuple, Dict
from collections.abc import Sequence

from xpra.util.objects import typedict
from xpra.codecs.image import ImageWrapper
from xpra.codecs.constants import get_subsampling_divs
from xpra.codecs.debug import may_save_image

from libc.string cimport memset  # pylint: disable=syntax-error
from xpra.buffers.membuf cimport getbuf, MemBuf  # pylint: disable=syntax-error
from xpra.codecs.avif.avif cimport (
    avifDecoder, avifResult, avifRGBImage, avifImage,
    avifResultToString,
    avifDecoderCreate, avifDecoderSetIOMemory,
    avifDecoderParse, avifDecoderNextImage, avifDecoderDestroy,
    avifDecoderNextImage, avifRGBImageSetDefaults, avifImageYUVToRGB,
    AVIF_RESULT, AVIF_RESULT_OK, AVIF_RGB_FORMAT_BGRA,
    AVIF_PIXEL_FORMAT_NONE, AVIF_PIXEL_FORMAT_YUV444, AVIF_PIXEL_FORMAT_YUV422,
    AVIF_PIXEL_FORMAT_YUV420, AVIF_PIXEL_FORMAT_YUV400,
    AVIF_RANGE_LIMITED, AVIF_RANGE_FULL,
    AVIF_VERSION_MAJOR, AVIF_VERSION_MINOR, AVIF_VERSION_PATCH,
    #AVIF_STRICT_ENABLED,
)
from xpra.buffers.membuf cimport memalign, buffer_context

from xpra.log import Logger
log = Logger("encoder", "avif")

from libc.stdint cimport uint8_t, uint32_t, uintptr_t
from libc.stdlib cimport free

cdef extern from *:
    ctypedef unsigned long size_t


AVIF_PIXEL_FORMAT: Dict[int, str] = {
    AVIF_PIXEL_FORMAT_NONE      : "NONE",
    AVIF_PIXEL_FORMAT_YUV444    : "YUV444",
    AVIF_PIXEL_FORMAT_YUV422    : "YUV422",
    AVIF_PIXEL_FORMAT_YUV420    : "YUV420",
    AVIF_PIXEL_FORMAT_YUV400    : "YUV400",
}

AVIF_RANGE: Dict[int, str] = {
    AVIF_RANGE_LIMITED  : "LIMITED",
    AVIF_RANGE_FULL     : "FULL",
}


def get_version() -> Tuple[int, int, int]:
    return (AVIF_VERSION_MAJOR, AVIF_VERSION_MINOR, AVIF_VERSION_PATCH)


def get_info() -> Dict:
    return  {
        "version"      : get_version(),
        "encodings"    : get_encodings(),
    }


def get_encodings() -> Sequence[str]:
    return ("avif", )


cdef inline void check(avifResult r, message: str):
    if r != AVIF_RESULT_OK:
        err = avifResultToString(r).decode("latin1") or AVIF_RESULT.get(r, r)
        raise RuntimeError("%s : %s" % (message, err))


def decompress_to_yuv(data: bytes, options: typedict) -> ImageWrapper:
    return decompress(data, options, True)


def decompress(data: bytes, options: typedict, yuv=False) -> ImageWrapper:
    cdef avifRGBImage rgb
    memset(&rgb, 0, sizeof(avifRGBImage))
    cdef avifDecoder * decoder = avifDecoderCreate()
    if decoder==NULL:
        raise RuntimeError("failed to create avif decoder")
    decoder.ignoreExif = 1
    decoder.ignoreXMP = 1
    #decoder.imageSizeLimit = 4096*4096
    #decoder.imageCountLimit = 1
    #decoder.strictFlags = AVIF_STRICT_ENABLED
    cdef avifResult r
    cdef size_t data_len
    cdef const uint8_t* data_buf
    cdef uint32_t width, height, stride, ydiv, size
    cdef uint8_t bpp = 32
    cdef MemBuf pixels
    cdef avifImage *image = NULL
    try:
        with buffer_context(data) as bc:
            data_len = len(bc)
            data_buf = <const uint8_t*> (<uintptr_t> int(bc))
            r = avifDecoderSetIOMemory(decoder, data_buf, data_len)
            check(r, "Cannot set IO on avifDecoder")

            r = avifDecoderParse(decoder)
            check(r, "Failed to decode image")

            image = decoder.image
            # Now available:
            # * All decoder->image information other than pixel data:
            # * width, height, depth
            # * transformations (pasp, clap, irot, imir)
            # * color profile (icc, CICP)
            # * metadata (Exif, XMP)
            # * decoder->alphaPresent
            # * number of total images in the AVIF (decoder->imageCount)
            # * overall image sequence timing (including per-frame timing with avifDecoderNthImageTiming())
            width = image.width
            height = image.height
            r = avifDecoderNextImage(decoder)
            check(r, "failed to get next image")
            # Now available (for this frame):
            # * All decoder->image YUV pixel data (yuvFormat, yuvPlanes, yuvRange, yuvChromaSamplePosition, yuvRowBytes)
            log("avif parsed: %ux%u (%ubpc) yuvFormat=%s, yuvRange=%s",
                width, height, image.depth, AVIF_PIXEL_FORMAT.get(image.yuvFormat), AVIF_RANGE.get(image.yuvRange))
            if yuv:
                #do we still need to copy if image.imageOwnsYUVPlanes?
                #could we keep the decoder context alive until the image is freed instead?
                yuv_format = "%sP" % AVIF_PIXEL_FORMAT.get(image.yuvFormat) #ie: YUV420P
                divs = get_subsampling_divs(yuv_format)
                planes = []
                strides = []
                for i in range(3):
                    ydiv = divs[i][1]
                    size = image.yuvRowBytes[i]*height//ydiv
                    planes.append(image.yuvPlanes[i][:size])
                    strides.append(image.yuvRowBytes[i])
                return ImageWrapper(0, 0, width, height, planes, yuv_format, 24, strides, planes=ImageWrapper.PLANAR_3)

            # * decoder->image alpha data (alphaRange, alphaPlane, alphaRowBytes)
            # * this frame's sequence timing
            avifRGBImageSetDefaults(&rgb, image)
            # Override YUV(A)->RGB(A) defaults here: depth, format, chromaUpsampling, ignoreAlpha, alphaPremultiplied, libYUVUsage, etc
            rgb_format = "BGRA"
            if not decoder.alphaPresent:
                rgb.ignoreAlpha = 1
                rgb_format = "BGRX"
                bpp = 24
            rgb.format = AVIF_RGB_FORMAT_BGRA
            stride = width*4
            pixels = getbuf(width*4*height, 0)
            rgb.pixels = <uint8_t *> pixels.get_mem()
            rgb.rowBytes = stride
            rgb.alphaPremultiplied = 1
            r = avifImageYUVToRGB(image, &rgb)
            check(r, "Conversion from YUV failed")
            if rgb.depth>8:
                raise ValueError("cannot handle depth %s" % rgb.depth)
            may_save_image("avif", data)
            if decoder.imageCount>1:
                log.warn("Warning: more than one image in avif data")
            img = ImageWrapper(0, 0, width, height, memoryview(pixels), rgb_format, bpp, stride, planes=ImageWrapper.PACKED)
            img.set_full_range(image.yuvRange == AVIF_RANGE_FULL and options.boolget("full-range", True))
            return img
    finally:
        avifDecoderDestroy(decoder)


def selftest(full=False) -> None:
    from xpra.codecs.checks import TEST_PICTURES
    for size, samples in TEST_PICTURES["avif"].items():
        w, h = size
        for bdata, options in samples:
            img = decompress(bdata, typedict(options))
            assert img.get_width()==w and img.get_height()==h
            assert len(img.get_pixels())>0
            #print("compressed data(%s)=%s" % (has_alpha, binascii.hexlify(r)))



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2024 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys


def main(args) -> int:
    if len(args) != 3:
        print("usage: %s host port" % (args[0], ))
        from xpra.exit_codes import ExitCode
        sys.exit(ExitCode.UNSUPPORTED)
    host = args[1]
    port = int(args[2])
    from xpra.client.qt6.client import run_client
    return run_client(host, port)


if __name__ == "__main__":
    main(sys.argv)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2022 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys
from PIL import Image

from xpra.util.objects import typedict
from xpra.util.env import envbool
from xpra.util.str_fn import memoryview_to_bytes
from xpra.codecs.image import ImageWrapper
from xpra.codecs.loader import load_codec


def main(files):
    assert len(files) > 0, "specify images to use for benchmark"
    avcodec = load_codec("dec_avcodec2")
    assert avcodec, "dec_avcodec is required"
    encoders = []
    ENCODERS = os.environ.get("XPRA_ENCODERS", "enc_vpx,enc_x264,nvenc").split(",")
    for encoder in ENCODERS:
        enc = load_codec(encoder)
        if not enc:
            print(f"{encoder} not found")
            continue
        print(f"* {encoder}")
        encodings = enc.get_encodings()
        for encoding in encodings:
            print(f"  - {encoding}")
            if encoding not in avcodec.get_encodings():
                print(f"    {avcodec} cannot decode {encoding}")
                continue
            matches = []
            for ics in enc.get_input_colorspaces(encoding):
                if ics not in ("BGRX", "YUV420P", "YUV444P",):
                    print(f"    skipping {ics}")
                    continue
                dcs = avcodec.get_output_colorspaces(encoding, ics)
                print(f"    dec_avcodec output colorspace for {encoding} + {ics} : {dcs}")
                if any(x in dcs for x in ("BGRX", "BGR", "YUV420P", "YUV444P", "GBRP")):
                    encoders.append((encoding, ics, enc))
                    matches.append(ics)
            if not matches:
                print(f"     no BGRX match for {encoding}")
    index = 0
    for f in files:
        index += 1
        img = Image.open(f)
        print(f"{index} : {f:40} : {img}")
        if img.mode != "RGBA":
            img = img.convert("RGBA")
        pixel_format = "BGRX"
        w, h = img.size
        rgb_data = img.tobytes("raw")
        stride = w * len(pixel_format)
        source_image = ImageWrapper(0, 0, w, h,
                                    rgb_data, pixel_format, len(pixel_format) * 8, stride,
                                    planes=ImageWrapper.PACKED, thread_safe=True)
        #source_image.restride(roundup(w*4, 8))
        for encoding, colorspace, enc in encoders:
            print(f"  {enc.get_type():10} {encoding:10} {colorspace}")
            image = source_image
            if colorspace != "BGRX":
                from xpra.codecs.libyuv.converter import Converter
                csc = Converter()
                csc.init_context(w, h, pixel_format,
                                 w, h, colorspace, typedict({"speed": 0}))
                image = csc.convert_image(source_image)
            encoder = enc.Encoder()
            try:
                encoder.init_context(encoding, w, h, image.get_pixel_format(), typedict({"quality": 100, "speed": 0}))
            except ValueError:
                print(f"  encoder rejected {w}x{h} {image.get_pixel_format()}")
                continue
            try:
                td = typedict()
                r = encoder.compress_image(image, td)
            except Exception:
                print(f"error on {enc.get_type()} : {enc.encode}")
                raise
            if not r:
                print(f"Error: no data for {enc.get_type()} : {enc.encode}")
                continue
            cdata, client_options = r
            print(f"r={r}")
            bdata = getattr(cdata, "data", cdata)
            if envbool("SAVE", False):
                filename = f"./{index}-{enc.get_type()}.{encoding.replace('/', '-')}"
                with open(filename, "wb") as fsave:
                    fsave.write(bdata)
            #now decode it back into an RGB picture:
            decoder = avcodec.Decoder()
            decoder.init_context(encoding, w, h, colorspace)
            decoded = decoder.decompress_image(bdata, typedict(client_options))
            dformat = decoded.get_pixel_format()
            output = decoded
            if dformat != "BGRX":
                from xpra.codecs.libyuv.converter import Converter
                csc = Converter()
                csc.init_context(w, h, dformat, w, h, "BGRX", typedict())
                output = csc.convert_image(decoded)
                print(f"    converted {dformat} to BGRX")
            obytes = memoryview_to_bytes(output.get_pixels())
            output_image = Image.frombuffer("RGBA", (w, h), obytes, "raw", "BGRA", output.get_rowstride())
            filename = f"{index}-{enc.get_type()}-{encoding}-{colorspace}.png"
            output_image.save(filename, "PNG")
            print(f"     saved to {filename!r}")


if __name__ == '__main__':
    assert len(sys.argv) > 1
    main(sys.argv[1:])



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any

from xpra.util.objects import typedict


class KeyboardConfigBase:
    """ Base class representing the keyboard configuration for a server.
    """
    __slots__ = ("enabled", "owner", "sync", "pressed_translation")

    def __init__(self):
        self.enabled = True
        self.owner = None
        self.sync = True
        self.pressed_translation: dict[int, tuple[int, int]] = {}

    def __repr__(self):
        return "KeyboardConfigBase"

    def get_info(self) -> dict[str, Any]:
        return {
            "enabled": self.enabled,
            "owner": self.owner or "",
            "sync": self.sync,
        }

    def parse_options(self, props: typedict) -> int:
        oldsync = self.sync
        keymap_dict = typedict(props.dictget("keymap") or {})
        self.sync = keymap_dict.boolget("sync", True)
        return int(oldsync != self.sync)

    def get_hash(self) -> str:
        return ""

    def set_layout(self, layout, variant, options) -> bool:
        """ should be overridden to configure the keyboard layout """
        return False

    def set_keymap(self, translate_only=False) -> None:
        """ should be overridden to configure the keymap """

    def set_default_keymap(self) -> None:
        """ should be overridden to set a default keymap """

    def make_keymask_match(self, modifier_list, ignored_modifier_keycode=None, ignored_modifier_keynames=None) -> None:
        """ should be overridden to match the modifier state specified """

    def get_keycode(self, client_keycode: int, keyname: str, pressed: bool,
                    modifiers, keyval: int, keystr: str, group: int) -> tuple[int, int]:
        if not keyname and client_keycode < 0:
            return -1, group
        if not pressed:
            r = self.pressed_translation.get(client_keycode)
            if r:
                # del self.pressed_translation[client_keycode]
                return r
        keycode, group = self.do_get_keycode(client_keycode, keyname, pressed, modifiers, keyval, keystr, group)
        if keycode < 0 and not keyname.islower():
            keyname = keyname.lower()
            keycode, group = self.do_get_keycode(client_keycode, keyname, pressed, modifiers, keyval, keystr, group)
        if pressed and keycode > 0:
            # keep track of it, so we can unpress the same key:
            self.pressed_translation[client_keycode] = keycode, group
        return keycode, group

    def do_get_keycode(self, client_keycode: int, keyname: str, pressed: bool,
                       modifiers, keyval: int, keystr: str, group: int) -> tuple[int, int]:
        from xpra.log import Logger  # pylint: disable=import-outside-toplevel
        log = Logger("keyboard")
        log("do_get_keycode%s", (client_keycode, keyname, pressed, modifiers, keyval, keystr, group))
        log.warn("Warning: %s does not implement get_keycode!", type(self))
        return -1, 0

    def is_modifier(self, _keycode: int) -> bool:
        # should be overridden in subclasses
        return False



# This file is part of Xpra.
# Copyright (C) 2022 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# duplicated from SimpleHTTPRequestHandler
# so that we can re-use it from the quic handler
# and also so that we can customize it more easily

import os
import sys
import html
from io import BytesIO
from urllib.parse import quote, unquote


def list_directory(path: str) -> tuple[int, dict[str, str], bytes]:
    try:
        dirlist = os.listdir(path)
    except OSError:
        return 404, {}, b"No permission to list directory"
    dirlist.sort(key=lambda a: a.lower())
    r = []
    try:
        displaypath = unquote(path, errors='surrogatepass')
    except UnicodeDecodeError:
        displaypath = unquote(path)
    displaypath = html.escape(displaypath, quote=False)
    enc = sys.getfilesystemencoding()
    title = f"Directory listing for {displaypath}"
    r.append('<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN" '
             '"http://www.w3.org/TR/html4/strict.dtd">')
    r.append('<html>\n<head>')
    r.append(f'<meta http-equiv="Content-Type" content="text/html; charset={enc}">')
    r.append(f'<title>{title}</title>\n</head>')
    r.append(f'<body>\n<h1>{title}</h1>')
    r.append('<hr>\n<ul>')
    for name in dirlist:
        fullname = os.path.join(path, name)
        displayname = linkname = name
        # Append / for directories or @ for symbolic links
        if os.path.isdir(fullname):
            displayname = name + "/"
            linkname = name + "/"
        if os.path.islink(fullname):
            displayname = name + "@"
            # Note: a link to a directory displays with @ and links with /
        href = quote(linkname, errors='surrogatepass')
        link = html.escape(displayname, quote=False)
        r.append(f'<li><a href="{href}">{link}</a></li>')
    r.append('</ul>\n<hr>\n</body>\n</html>\n')
    encoded = '\n'.join(r).encode(enc, 'surrogateescape')
    f = BytesIO()
    f.write(encoded)
    contents = f.getvalue()
    f.close()
    return 200, {
        "Content-type": f"text/html; charset={enc}",
        "Content-Length": str(len(encoded)),
    }, contents
